/*
** Beispiel: Erzeuger/Verbraucher-Problem als Monitor mit Mutex und
** Bedingungsvariablen realisiert.
** Uebersetzung des Programms: 
**     javac -classpath ../BSsync.jar ErzVerb.java
** Start des Programms:
**     java -classpath ../BSsync.jar:. ErzVerb
*/

import BSsync.Lock;
import BSsync.Condition;

/*
**************************************************************************
** Realisierung des beschraenkten Puffers durch ein Array (Ringpuffer)
**  - wr: Schreibposition
**  - rd: Leseposition
*/

class Buffer {
	private int[] buf;
	private int wr;
	private int rd;
	private int size;
	
	public Buffer(int size)
	{
		buf = new int[size];
		wr = 0;
		rd = 0;
		this.size = size;
	}
		
	public void insertItem(int item)
	{
		buf[wr] = item;
		wr = (wr + 1) % size;
	}
	
	public int removeItem()
	{
		int res = buf[rd];
		rd = (rd + 1) % size;
		return res;
	}

	public int getSize()
	{
		return size;
	}
}

/*
**************************************************************************
** Klassen fuer die Erzeuger- und Verbraucher-Threads
*/

class Producer extends Thread
{
	private ErzVerb ev;

	public Producer(ErzVerb ev)
	{
		this.ev = ev;
	}

	public void run()
	{
		int item = 0;
		int i;
		for (i=0; i<100; i++) {
			item++;               // Erzeuge item ...
			ev.Insert(item);
		}
	}
}

class Consumer extends Thread
{
	private ErzVerb ev;

	public Consumer(ErzVerb ev)
	{
		this.ev = ev;
	}

	public void run()
	{
		int item;
		int i;
		for (i=0; i<100; i++) {
			item = ev.Remove();
			System.out.println(item);  // Verbrauche item ...
		}
	}
}

public class ErzVerb {

	/*
    **************************************************************************
	** Monitor fuer Erzeuger/Verbraucher-Problem
	*/
	
	/*
	** Mutex fuer wechselseitigen Ausschluss, zwei Bedingungsvariable:
	**  - full: verhindert Schreiben in vollen Puffer
	**  - empty: verhindert Lesen aus leerem Puffer
	** Variable count zaehlt belegte Pufferplaetze
	*/
	private Lock mutex;
	private Condition full;
	private Condition empty;
	private int count;

	Buffer buffer;

	public ErzVerb(int size)
	{
		buffer = new Buffer(size);
		mutex = new Lock("mutex");
		full = mutex.newCondition("full");
		empty = mutex.newCondition("empty");
		count = 0;
	}
		
	/*
	** Einfuegeprozedur
	*/
	public void Insert(int item)
	{
		mutex.lock();
		while (count == buffer.getSize())
			full.await();
		buffer.insertItem(item);
		count++;
		if (count == 1)
			empty.signal();
		mutex.unlock();
	}

	/*
	** Entfernprozedur
	*/
	public int Remove()
	{
		int result;
		mutex.lock();
		while (count == 0)
			empty.await();
		result = buffer.removeItem();
		count--;
		if (count == buffer.getSize()-1)
			full.signal();
		mutex.unlock();
		return result;
	}

	/*
	** Hauptprogramm
	*/
	public static void main(String argv[])
	{
		ErzVerb ev = new ErzVerb(5);

		Lock.verbose = true;

		/*
		** Erzeuge je einen Thread fuer die Prozedured Producer und Consumer
		*/
		Producer prod = new Producer(ev);
		Consumer cons = new Consumer(ev);

		prod.start();
		cons.start();
	}
}

