/*
** Beispiel: Erzeuger/Verbraucher-Problem mit Semaphoren realisiert.
** Uebersetzung des Programms: 
**     javac -classpath ../BSsync.jar ErzVerb.java
** Start des Programms:
**     java -classpath ../BSsync.jar:. ErzVerb
*/

import BSsync.Semaphore;

/*
**************************************************************************
** Realisierung des beschraenkten Puffers durch ein Array (Ringpuffer)
**  - wr: Schreibposition
**  - rd: Leseposition
*/

class Buffer {
	private int[] buf;
	private int wr;
	private int rd;
	private int size;
	
	public Buffer(int size)
	{
		buf = new int[size];
		wr = 0;
		rd = 0;
		this.size = size;
	}
		
	public void insertItem(int item)
	{
		buf[wr] = item;
		wr = (wr + 1) % size;
	}
	
	public int removeItem()
	{
		int res = buf[rd];
		rd = (rd + 1) % size;
		return res;
	}

	public int getSize()
	{
		return size;
	}
}

/*
**************************************************************************
** Klassen fuer die Erzeuger- und Verbraucher-Threads
*/

class Producer extends Thread
{
	private ErzVerb ev;

	public Producer(ErzVerb ev)
	{
		this.ev = ev;
	}

	public void run()
	{
		int item = 0;
		int i;
		for (i=0; i<1000; i++) {
			item++;               // Erzeuge item ...
			ev.Insert(item);
		}
	}
}

class Consumer extends Thread
{
	private ErzVerb ev;

	public Consumer(ErzVerb ev)
	{
		this.ev = ev;
	}

	public void run()
	{
		int item;
		int i;
		for (i=0; i<1000; i++) {
			item = ev.Remove();
			System.out.println(item);  // Verbrauche item ...
		}
	}
}

public class ErzVerb {

	/*
    **************************************************************************
	** Monitor fuer Erzeuger/Verbraucher-Problem
	*/
	
	/*
	** Semaphore:
	**  - mutex: fuer wechselseitigen Ausschluss
	**  - full: zaehlt volle Pufferplaetze
	**  - empty: zaehlt leere Pufferplaetze
	** Variable count zaehlt belegte Pufferplaetze
	*/
	private Semaphore mutex;
	private Semaphore full;
	private Semaphore empty;

	Buffer buffer;

	public ErzVerb(int size)
	{
		buffer = new Buffer(size);
		mutex = new Semaphore(1, "mutex");
		full = new Semaphore(0, "full");
		empty = new Semaphore(size, "empty");
	}
		
	/*
	** Einfuegeprozedur
	*/
	public void Insert(int item)
	{
		empty.acquire();
		mutex.acquire();
		buffer.insertItem(item);
		mutex.release();
		full.release();
	}

	/*
	** Entfernprozedur
	*/
	public int Remove()
	{
		int result;
		full.acquire();
		mutex.acquire();
		result = buffer.removeItem();
		mutex.release();
		empty.release();
		return result;
	}

	/*
	** Hauptprogramm
	*/
	public static void main(String argv[])
	{
		ErzVerb ev = new ErzVerb(5);

		Semaphore.verbose = true;

		/*
		** Erzeuge je einen Thread fuer die Prozedured Producer und Consumer
		*/
		Producer prod = new Producer(ev);
		Consumer cons = new Consumer(ev);

		prod.start();
		cons.start();
	}
}

